defmodule Instream.Series.Hydrator do
  @moduledoc false

  @doc """
  Converts a plain map into a series definition struct.

  Keys not defined in the series are silently dropped.
  """
  @spec from_map(module, map) :: struct
  def from_map(series, data) do
    data_fields = Map.take(data, series.__meta__(:fields))
    data_tags = Map.take(data, series.__meta__(:tags))

    struct(series, %{
      fields: struct(Module.concat(series, Fields), data_fields),
      tags: struct(Module.concat(series, Tags), data_tags),
      timestamp: convert_to_timestamp(data[:time] || data[:timestamp])
    })
  end

  @doc """
  Converts a query result map into a list of series definition structs.

  Keys not defined in the series are silently dropped.
  """
  @spec from_result(module, map) :: [struct]
  def from_result(series, %{results: [%{series: [data]}]}) do
    data_tags = Map.get(data, :tags, %{})

    Enum.map(Map.get(data, :values, []), fn values ->
      data_fields =
        data.columns
        |> Enum.zip(values)
        |> Enum.into(%{}, fn {k, v} -> {String.to_atom(k), v} end)

      from_map(series, Map.merge(data_tags, data_fields))
    end)
  end

  defp convert_to_timestamp(time) when is_integer(time), do: time

  # `DateTime` only supports :microsecond level precisions
  # OTP 21.0 is required for full precision
  if Code.ensure_loaded?(:calendar) && function_exported?(:calendar, :rfc3339_to_system_time, 2) do
    defp convert_to_timestamp(time) when is_binary(time) do
      time
      |> String.to_charlist()
      |> :calendar.rfc3339_to_system_time(unit: :nanosecond)
    rescue
      _ -> nil
    end
  else
    defp convert_to_timestamp(time) when is_binary(time) do
      case DateTime.from_iso8601(time) do
        {:ok, datetime, 0} -> DateTime.to_unix(datetime, :nanosecond)
        _ -> nil
      end
    end
  end

  defp convert_to_timestamp(_), do: nil
end
